@@ -219,8 +219,14 @@ int do_backup(Options *op, const char *filename)
     }
     
     if (lstat(filename, &stat_buf) == -1) {
-        ui_error(op, "Unable to determine properties for file '%s' (%s).",
-                 filename, strerror(errno));
+        switch (errno) {
+        case ENOENT:
+            ret_val = TRUE;
+            break;
+        default:
+            ui_error(op, "Unable to determine properties for file '%s' (%s).",
+                     filename, strerror(errno));
+        }
         goto done;
     }
 
@@ -509,7 +515,7 @@ static int do_uninstall(Options *op)
                     existing_installation_is_borked, op->log_file_name);
 
         } else {
-            ui_warn(op, existing_installation_is_borked);
+            ui_warn(op, "%s", existing_installation_is_borked);
         }
     }
     
@@ -612,14 +618,14 @@ static int do_uninstall(Options *op)
             if (!nvrename(op, tmpstr, e->filename)) {
                 ui_warn(op, "Unable to restore file '%s'.", e->filename);
             } else {
-                if (chmod(e->filename, e->mode) == -1) {
-                    ui_warn(op, "Unable to restore permissions %04o for "
-                            "file '%s'.", e->mode, e->filename);
+                if (chown(e->filename, e->uid, e->gid)) {
+                    ui_warn(op, "Unable to restore owner (%d) and group "
+                            "(%d) for file '%s' (%s).",
+                            e->uid, e->gid, e->filename, strerror(errno));
                 } else {
-                    if (chown(e->filename, e->uid, e->gid)) {
-                        ui_warn(op, "Unable to restore owner (%d) and group "
-                                "(%d) for file '%s' (%s).",
-                                e->uid, e->gid, e->filename, strerror(errno));
+                    if (chmod(e->filename, e->mode) == -1) {
+                        ui_warn(op, "Unable to restore permissions %04o for "
+                                "file '%s'.", e->mode, e->filename);
                     }
                 }
             }
@@ -653,7 +659,7 @@ static BackupInfo *read_backup_log_file(Options *op)
 {
     struct stat stat_buf;
     char *buf, *c, *line, *filename;
-    int fd, num, line_num = 0;
+    int fd, num, length, line_num = 0;
     float percent;
     
     BackupLogEntry *e;
@@ -692,7 +698,9 @@ static BackupInfo *read_backup_log_file(Options *op)
 
     /* map the file */
 
-    buf = mmap(0, stat_buf.st_size, PROT_READ, MAP_FILE | MAP_SHARED, fd, 0);
+    length = stat_buf.st_size;
+
+    buf = mmap(0, length, PROT_READ, MAP_FILE | MAP_SHARED, fd, 0);
     if (!buf) {
         ui_error(op, "Unable to mmap file '%s' (%s).", strerror(errno));
         return NULL;
@@ -702,13 +710,13 @@ static BackupInfo *read_backup_log_file(Options *op)
 
     b = nvalloc(sizeof(BackupInfo));
 
-    b->version = get_next_line(buf, &c);
+    b->version = get_next_line(buf, &c, buf, length);
     if (!b->version || !c) goto parse_error;
     
     percent = (float) (c - buf) / (float) stat_buf.st_size;
     ui_status_update(op, percent, NULL);
 
-    b->description = get_next_line(c, &c);
+    b->description = get_next_line(c, &c, buf, length);
     if (!b->description || !c) goto parse_error;
 
     b->n = 0;
@@ -722,7 +730,7 @@ static BackupInfo *read_backup_log_file(Options *op)
         
         /* read and parse the next line */
 
-        line = get_next_line(c, &c);
+        line = get_next_line(c, &c, buf, length);
         if (!line) break;
 
         if (!parse_first_line(line, &num, &filename)) goto parse_error;
@@ -743,7 +751,8 @@ static BackupInfo *read_backup_log_file(Options *op)
         switch(e->num) {
 
         case INSTALLED_FILE:
-            if ((line = get_next_line(c, &c)) == NULL) goto parse_error;
+            line = get_next_line(c, &c, buf, length);
+            if (line == NULL) goto parse_error;
             line_num++;
 
             if (!parse_crc(line, &e->crc)) goto parse_error;
@@ -752,7 +761,8 @@ static BackupInfo *read_backup_log_file(Options *op)
             break;
 
         case INSTALLED_SYMLINK:
-            if ((line = get_next_line(c, &c)) == NULL) goto parse_error;
+            line = get_next_line(c, &c, buf, length);
+            if (line == NULL) goto parse_error;
             line_num++;
             
             e->target = line;
@@ -760,12 +770,14 @@ static BackupInfo *read_backup_log_file(Options *op)
             break;
             
         case BACKED_UP_SYMLINK:
-            if ((line = get_next_line(c, &c)) == NULL) goto parse_error;
+            line = get_next_line(c, &c, buf, length);
+            if (line == NULL) goto parse_error;
             line_num++;
             
             e->target = line;
 
-            if ((line = get_next_line(c, &c)) == NULL) goto parse_error;
+            line = get_next_line(c, &c, buf, length);
+            if (line == NULL) goto parse_error;
             line_num++;
 
             if (!parse_mode_uid_gid(line, &e->mode, &e->uid, &e->gid))
@@ -777,7 +789,8 @@ static BackupInfo *read_backup_log_file(Options *op)
         default:
             if (num < BACKED_UP_FILE_NUM) goto parse_error;
             
-            if ((line = get_next_line(c, &c)) == NULL) goto parse_error;
+            line = get_next_line(c, &c, buf, length);
+            if (line == NULL) goto parse_error;
             line_num++;
 
             if (!parse_crc_mode_uid_gid(line, &e->crc, &e->mode,
@@ -974,23 +987,25 @@ char *get_installed_driver_version_and_descr(Options *op, int *major,
 {
     struct stat stat_buf;
     char *c, *version = NULL, *descr = NULL, *buf = NULL;
-    int fd = -1;
+    int length, fd = -1;
 
     if ((fd = open(BACKUP_LOG, O_RDONLY)) == -1) goto done;
     
     if (fstat(fd, &stat_buf) == -1) goto done;
     
     /* map the file */
+
+    length = stat_buf.st_size;
     
-    buf = mmap(0, stat_buf.st_size, PROT_READ, MAP_FILE | MAP_SHARED, fd, 0);
+    buf = mmap(0, length, PROT_READ, MAP_FILE | MAP_SHARED, fd, 0);
     if (!buf) goto done;
     
-    version = get_next_line(buf, &c);
+    version = get_next_line(buf, &c, buf, length);
     if (!version) goto done;
 
     if (!nvid_version(version, major, minor, patch)) goto done;
     
-    descr = get_next_line(c, NULL);
+    descr = get_next_line(c, NULL, buf, length);
 
  done:
     if (version) free(version);
@@ -1011,9 +1026,9 @@ char *get_installed_driver_version_and_descr(Options *op, int *major,
  * Returns FALSE if the user decided they didn't want to continue with
  * installation.
  *
- * XXX what happens if the existing driver wasn't installed with this
- * installation mechanism?  If we can't find BACKUP_LOG, should we
- * look in the usual places to see if our driver files are present?
+ * If we are only installing a kernel module, then there must be an
+ * existing driver installation, and the version of that installation
+ * must match the module we're trying to install.
  */
 
 int check_for_existing_driver(Options *op, Package *p)
@@ -1024,8 +1039,33 @@ int check_for_existing_driver(Options *op, Package *p)
     if (!check_for_existing_rpms(op)) return FALSE;
 
     descr = get_installed_driver_version_and_descr(op, &major, &minor, &patch);
-    if (!descr) return TRUE;
+
+    if (op->kernel_module_only) {
+        if (!descr) {
+            ui_error(op, "No NVIDIA driver is currently installed; the "
+                     "'--kernel-module-only' option can only be used "
+                     "to install the NVIDIA kernel module on top of an "
+                     "existing driver installation.");
+            return FALSE;
+        } else {
+            if ((p->major != major) ||
+                (p->minor != minor) ||
+                (p->patch != patch)) {
+                ui_error(op, "The '--kernel-module-only' option can only be "
+                         "used to install a kernel module on top of an "
+                         "existing driver installation of the same driver "
+                         "version.  The existing driver installation is "
+                         "%d.%d-%d, but the kernel module is %d.%d-%d\n",
+                         major, minor, patch, p->major, p->minor, p->patch);
+                return FALSE;
+            } else {
+                return TRUE;
+            }
+        }
+    }
     
+    if (!descr) return TRUE;
+
     /*
      * XXX we could do a comparison, here, to check that the Package
      * version is greater than or equal to the installed version, and
@@ -1144,6 +1184,38 @@ int test_installed_files(Options *op)
 
 
 
+/*
+ * find_installed_file() - scan the backup log file for the specified
+ * filename; return TRUE if the filename is listed as an installed file.
+ */
+
+int find_installed_file(Options *op, char *filename)
+{
+    BackupInfo *b;
+    BackupLogEntry *e;
+    int i;
+    
+    if ((b = read_backup_log_file(op)) == NULL) return FALSE;
+
+    for (i = 0; i < b->n; i++) {
+        e = &b->e[i];
+        if ((e->num == INSTALLED_FILE) && strcmp(filename, e->filename) == 0) {
+            
+            /* XXX should maybe compare inodes rather than
+               filenames? */
+
+            return TRUE;
+        }
+    }
+
+    /* XXX need to free b */
+
+    return FALSE;
+
+} /* find_installed_file() */
+
+
+
 /*
  * sanity_check_backup_log_entries() - this function is very similar
  * to check_backup_log_entries(); however, it varies in it's error
