@@ -48,13 +48,13 @@
 #include "backup.h"
 #include "files.h"
 #include "misc.h"
+#include "sanity.h"
 
 /* local prototypes */
 
 
 static Package *parse_manifest(Options *op);
-
-
+static int install_kernel_module(Options *op,  Package *p);
 
 
 /*
@@ -78,11 +78,12 @@ int install_from_cwd(Options *op)
     Package *p;
     CommandList *c;
     const char *msg;
+    int ret;
 
     static const char edit_your_xf86config[] =
-        "Please update your XF86Config file as "
+        "Please update your XF86Config or xorg.conf file as "
         "appropriate; see the file /usr/share/doc/"
-        "NVIDIA_GLX-1.0/README for details.";
+        "NVIDIA_GLX-1.0/README.txt for details.";
 
     static const char suse_edit_your_xf86config[] =
         "On SuSE Linux/United Linux please use SaX2 now to enable "
@@ -98,14 +99,21 @@ int install_from_cwd(Options *op)
     ui_set_title(op, "%s (%d.%d-%d)", p->description,
                  p->major, p->minor, p->patch);
 
-    /* make sure the kernel module is unloaded */
-    
-    if (!check_for_unloaded_kernel_module(op, p)) goto failed;
-    
+    /* 
+     * warn the user if "legacy" GPUs are installed in this system
+     * and if no supported GPU is found, at all.
+     */
+
+    check_for_nvidia_graphics_devices(op, p);
+
     /* check that we are not running any X server */
 
     if (!check_for_running_x(op)) goto failed;
 
+    /* make sure the kernel module is unloaded */
+    
+    if (!check_for_unloaded_kernel_module(op, p)) goto failed;
+    
     /* ask the user to accept the license */
     
     if (!get_license_acceptance(op)) return FALSE;
@@ -116,10 +124,196 @@ int install_from_cwd(Options *op)
      */
 
     if (!check_for_existing_driver(op, p)) return FALSE;
+
+    /* attempt to build a kernel module for the target kernel */
+
+    if (!op->no_kernel_module) {
+        if (!install_kernel_module(op, p)) goto failed;
+    } else {
+        ui_warn(op, "You specified the '--no-kernel-module' command line "
+                "option, nvidia-installer will not install a kernel "
+                "module as part of this driver installation, and it will "
+                "not remove existing NVIDIA kernel modules not part of "
+                "an earlier NVIDIA driver installation.  Please ensure "
+                "that an NVIDIA kernel module matching this driver version "
+                "is installed seperately.");
+    }
     
+    /*
+     * if we are only installing the kernel module, then remove
+     * everything else from the package; otherwise do some
+     * OpenGL-specific stuff
+     */
+
+    if (op->kernel_module_only) {
+        remove_non_kernel_module_files_from_package(op, p);
+    } else {
+
+        /* ask for the XFree86 and OpenGL installation prefixes. */
+    
+        if (!get_prefixes(op)) goto failed;
+
+        /* ask if we should install the OpenGL header files */
+
+        should_install_opengl_headers(op, p);
+
+        /*
+         * select the appropriate TLS class, modifying the package as
+         * necessary.
+         */
+    
+        select_tls_class(op, p);
+
+        /*
+         * if the package contains any libGL.la or .desktop files,
+         * process them (perform some search and replacing so
+         * that they reflect the correct installation path, etc)
+         * and add them to the package list (files to be installed).
+         */
+        
+        process_libGL_la_files(op, p);
+        process_dot_desktop_files(op, p);
+
+#if defined(NV_X86_64)
+        /*
+         * ask if we should install the 32bit compatibility files on
+         * this machine.
+         */
+
+        should_install_compat32_files(op, p);
+#endif /* NV_X86_64 */
+    }
+
+    /*
+     * now that we have the installation prefixes, build the
+     * destination for each file to be installed
+     */
+    
+    if (!set_destinations(op, p)) goto failed;
+    
+    /*
+     * uninstall the existing driver; this needs to be done before
+     * building the command list.
+     *
+     * XXX if we uninstall now, then build the command list, and
+     * then ask the user if they really want to execute the
+     * command list, if the user decides not to execute the
+     * command list, they'll be left with no driver installed.
+     */
+
+    if (!op->kernel_module_only) {
+        if (!uninstall_existing_driver(op, FALSE)) goto failed;
+    }
+
+    /* build a list of operations to execute to do the install */
+    
+    if ((c = build_command_list(op, p)) == NULL) goto failed;
+
+    /* call the ui to get approval for the list of commands */
+    
+    if (!ui_approve_command_list(op, c, "%s", p->description)) return FALSE;
+    
+    /* initialize the backup log file */
+
+    if (!op->kernel_module_only) {
+        if (!init_backup(op, p)) goto failed;
+    }
+
+    /* execute the command list */
+
+    if (!do_install(op, p, c)) goto failed;
+  
+    /*
+     * check that everything is installed properly (post-install
+     * sanity check)
+     */
+
+    check_installed_files_from_package(op, p);
+
+    if (!check_sysvipc(op)) goto failed;
+    if (!check_runtime_configuration(op, p)) goto failed;
+    
+    /* done */
+
+    if (op->kernel_module_only) {
+        ui_message(op, "Installation of the kernel module for the %s "
+                   "(version %s) is now complete.",
+                   p->description, p->version_string);
+    } else {
+        
+        /* ask the user if they would like to run nvidia-xconfig */
+        
+        ret = ui_yes_no(op, op->run_nvidia_xconfig,
+                        "Would you like to run the nvidia-xconfig utility "
+                        "to automatically update your X configuration file "
+                        "so that the NVIDIA X driver will be used when you "
+                        "restart X?  Any pre-existing X configuration "
+                        "file will be backed up.");
+        
+        if (ret) {
+            ret = run_nvidia_xconfig(op);
+        }
+        
+        if (ret) {
+            ui_message(op, "Your X configuration file has been successfully "
+                       "updated.  Installation of the %s (version: %s) is now "
+                       "complete.", p->description, p->version_string);
+        } else {
+            
+            if ((op->distro == SUSE) || (op->distro == UNITED_LINUX)) {
+                msg = suse_edit_your_xf86config;
+            } else {
+                msg = edit_your_xf86config;
+            }
+            
+            ui_message(op, "Installation of the %s (version: %s) is now "
+                       "complete.  %s", p->description,
+                       p->version_string, msg);
+        }
+    }
+    
+    return TRUE;
+    
+ failed:
+
+    if (op->logging) {
+        ui_error(op, "Installation has failed.  Please see the file '%s' "
+                 "for details.  You may find suggestions on fixing "
+                 "installation problems in the README available on the "
+                 "Linux driver download page at www.nvidia.com.",
+                 op->log_file_name);
+    } else {
+        ui_error(op, "Installation has failed.  You may find suggestions "
+                 "on fixing installation problems in the README available "
+                 "on the Linux driver download page at www.nvidia.com.");
+    }
+    
+    return FALSE;
+
+} /* install_from_cwd() */
+
+
+
+/*
+ * install_kernel_module() - attempt to build and install a kernel
+ * module for the running kernel; we first check if a prebuilt kernel
+ * interface file exists. If yes, we try to link it into the final
+ * kernel module, else we try to build one from source.
+ *
+ * If we succeed in building a kernel module, we attempt to load it
+ * into the host kernel and add it to the list of files to install if
+ * the load attempt succeeds.
+ */
+
+static int install_kernel_module(Options *op,  Package *p)
+{
     /* determine where to install the kernel module */
     
-    if (!determine_kernel_module_installation_path(op)) goto failed;
+    if (!determine_kernel_module_installation_path(op)) return FALSE;
+
+    /* check '/proc/sys/kernel/modprobe' */
+
+    if (!check_proc_modprobe_path(op)) return FALSE;
 
     /*
      * do nvchooser-style logic to decide if we have a prebuilt kernel
@@ -146,23 +340,34 @@ int install_from_cwd(Options *op)
          * abort.
          */
         
-        if (!link_kernel_module(op, p)) goto failed;
+        if (!link_kernel_module(op, p)) return FALSE;
 
     } else {
-        
-        /* XXX we need to do a cc version check */
+        /*
+         * make sure the required development tools are present on
+         * this system before attempting to verify the compiler and
+         * trying to build a custom kernel interface.
+         */
+        if (!check_development_tools(op, p)) return FALSE;
+
+        /*
+         * make sure that the selected or default system compiler
+         * is compatible with the target kernel; the user may choose
+         * to override the check.
+         */
+        if (!check_cc_version(op, p)) return FALSE;
 
         /*
          * we do not have a prebuilt kernel interface; thus we'll need
          * to compile the kernel interface, so determine where the
-         * kernel header files are
+         * kernel source files are.
          */
         
-        if (!determine_kernel_include_path(op)) goto failed;
+        if (!determine_kernel_source_path(op, p)) return FALSE;
     
         /* and now, build the kernel interface */
         
-        if (!build_kernel_module(op, p)) goto failed;
+        if (!build_kernel_module(op, p)) return FALSE;
     }
 
     /*
@@ -170,104 +375,14 @@ int install_from_cwd(Options *op)
      * to be sure it's OK
      */
     
-    if (!test_kernel_module(op, p)) goto failed;
+    if (!test_kernel_module(op, p)) return FALSE;
     
     /* add the kernel module to the list of things to install */
     
-    if (!add_kernel_module_to_package(op, p)) goto failed;
-    
-    /* ask for the XFree86 and OpenGL installation prefixes. */
-    
-    if (!get_prefixes(op)) goto failed;
-
-    /* ask if we should install the OpenGL header files */
-
-    should_install_opengl_headers(op, p);
-
-    /*
-     * select the appropriate TLS class, modifying the package as
-     * necessary.
-     */
-    
-    select_tls_class(op, p);
-    
-    /*
-     * now that we have the installation prefixes, build the
-     * destination for each file to be installed
-     */
-
-    if (!set_destinations(op, p)) goto failed;
-    
-    /*
-     * uninstall the existing driver; this needs to be done before
-     * building the command list.
-     *
-     * XXX if we uninstall now, then build the command list, and then
-     * ask the user if they really want to execute the command list,
-     * if the user decides not to execute the command list, they'll be
-     * left with no driver installed.
-     */
-
-    if (!uninstall_existing_driver(op, FALSE)) goto failed;
-
-    /* build a list of operations to execute to do the install */
-    
-    if ((c = build_command_list(op, p)) == NULL) goto failed;
-
-    /* call the ui to get approval for the list of commands */
-    
-    if (!ui_approve_command_list(op, c, p->description)) return FALSE;
-    
-    /* initialize the backup log file */
-
-    if (!init_backup(op, p)) goto failed;
-
-    /* execute the command list */
-
-    if (!do_install(op, p, c)) goto failed;
-
-    /* 
-     * XXX IMPLEMENT ME: generate an XF86Config file?
-     */
-  
-    /*
-     * check that everything is installed properly (post-install
-     * sanity check)
-     */
-
-    check_installed_files_from_package(op, p);
-
-    /* done */
-
-    if ((op->distro == SUSE) || (op->distro == UNITED_LINUX)) {
-        msg = suse_edit_your_xf86config;
-    } else {
-        msg = edit_your_xf86config;
-    }
-    
-    ui_message(op, "Installation of the %s (version: %s) is now "
-               "complete.  %s", p->description, p->version_string, msg);
+    if (!add_kernel_module_to_package(op, p)) return FALSE;
     
     return TRUE;
-    
- failed:
-
-    if (op->logging) {
-        ui_error(op, "Installation has failed.  Please see the file '%s' "
-                 "for details.  You may find suggestions on fixing "
-                 "installation problems in the README available on the "
-                 "Linux driver download page at www.nvidia.com.",
-                 op->log_file_name);
-    } else {
-        ui_error(op, "Installation has failed.  You may find suggestions "
-                 "on fixing installation problems in the README available "
-                 "on the Linux driver download page at www.nvidia.com.");
-    }
-    
-    return FALSE;
-
-} /* install_from_cwd() */
-
+}
 
 
 /*
@@ -286,7 +401,7 @@ int add_this_kernel(Options *op)
 
     /* find the kernel header files */
 
-    if (!determine_kernel_include_path(op)) goto failed;
+    if (!determine_kernel_source_path(op, p)) goto failed;
 
     /* build the precompiled kernel interface */
 
@@ -333,6 +448,8 @@ int add_this_kernel(Options *op)
  *   - a filename (relative to the cwd)
  *   - an octal value describing the permissions
  *   - a flag describing the file type
+ *   - certain file types have an architecture
+ *   - certain file types have a second flag
  *   - certain file types will have a path
  *   - symbolic links will name the target of the link
  */
@@ -344,7 +461,7 @@ static Package *parse_manifest (Options *op)
     int fd, len = 0;
     struct stat stat_buf;
     Package *p;
-    char *manifest = NULL, *ptr;
+    char *manifest = MAP_FAILED, *ptr;
     
     p = (Package *) nvalloc(sizeof (Package));
     
@@ -362,47 +479,41 @@ static Package *parse_manifest (Options *op)
     len = stat_buf.st_size;
 
     manifest = mmap(0, len, PROT_READ, MAP_FILE|MAP_SHARED, fd, 0);
-    if (manifest == (void *) -1) goto cannot_open;
+    if (manifest == MAP_FAILED) goto cannot_open;
     
     /* the first line is the description */
 
     line = 1;
-    p->description = get_next_line(manifest, &ptr);
+    p->description = get_next_line(manifest, &ptr, manifest, len);
     if (!p->description) goto invalid_manifest_file;
     
     /* the second line is the version */
     
     line++;
-    p->version_string = get_next_line(ptr, &ptr);
+    p->version_string = get_next_line(ptr, &ptr, manifest, len);
     if (!p->version_string) goto invalid_manifest_file;
     if (!nvid_version(p->version_string, &p->major, &p->minor, &p->patch))
         goto invalid_manifest_file;
     
-    /* the third line is the kernel module filename */
+    /* new third line is the kernel interface filename */
 
     line++;
-    p->kernel_module_filename = get_next_line(ptr, &ptr);
-    if (!p->kernel_module_filename) goto invalid_manifest_file;
+    p->kernel_interface_filename = get_next_line(ptr, &ptr, manifest, len);
+    if (!p->kernel_interface_filename) goto invalid_manifest_file;
 
-    /* new fourth line is the kernel interface filename */
+    /* the fourth line is the kernel module name */
 
     line++;
-    p->kernel_interface_filename = get_next_line(ptr, &ptr);
-    if (!p->kernel_module_filename) goto invalid_manifest_file;
-
-    /* the fifth line is the kernel module name */
-
-    line++;
-    p->kernel_module_name = get_next_line(ptr, &ptr);
+    p->kernel_module_name = get_next_line(ptr, &ptr, manifest, len);
     if (!p->kernel_module_name) goto invalid_manifest_file;
 
     /*
-     * the sixth line is a whitespace-separated list of kernel modules
+     * the fifth line is a whitespace-separated list of kernel modules
      * to be unloaded before installing the new kernel module
      */
 
     line++;
-    tmpstr = get_next_line(ptr, &ptr);
+    tmpstr = get_next_line(ptr, &ptr, manifest, len);
     if (!tmpstr) goto invalid_manifest_file;
 
     p->bad_modules = NULL;
@@ -417,13 +528,13 @@ static Package *parse_manifest (Options *op)
     } while (p->bad_modules[n-1]);
     
     /*
-     * the seventh line is a whitespace-separated list of kernel module
+     * the sixth line is a whitespace-separated list of kernel module
      * filenames to be uninstalled before installing the new kernel
      * module
      */
 
     line++;
-    tmpstr = get_next_line(ptr, &ptr);
+    tmpstr = get_next_line(ptr, &ptr, manifest, len);
     if (!tmpstr) goto invalid_manifest_file;
 
     p->bad_module_filenames = NULL;
@@ -437,20 +548,21 @@ static Package *parse_manifest (Options *op)
         p->bad_module_filenames[n-1] = read_next_word(c, &c);
     } while (p->bad_module_filenames[n-1]);
     
-    /* the eighth line is the kernel module build directory */
+    /* the seventh line is the kernel module build directory */
 
     line++;
-    p->kernel_module_build_directory = get_next_line(ptr, &ptr);
+    p->kernel_module_build_directory = get_next_line(ptr, &ptr, manifest, len);
     if (!p->kernel_module_build_directory) goto invalid_manifest_file;
     remove_trailing_slashes(p->kernel_module_build_directory);
 
     /*
-     * the nineth line is the directory containing precompiled kernel
+     * the eigth line is the directory containing precompiled kernel
      * interfaces
      */
 
     line++;
-    p->precompiled_kernel_interface_directory = get_next_line(ptr, &ptr);
+    p->precompiled_kernel_interface_directory =
+        get_next_line(ptr, &ptr, manifest, len);
     if (!p->precompiled_kernel_interface_directory)
         goto invalid_manifest_file;
     remove_trailing_slashes(p->precompiled_kernel_interface_directory);
@@ -461,7 +573,7 @@ static Package *parse_manifest (Options *op)
     line++;
     
     do {
-        buf = get_next_line(ptr, &ptr);
+        buf = get_next_line(ptr, &ptr, manifest, len);
         if ((!buf) || (buf[0] == '\0')) {
             done = TRUE;
         } else {
@@ -496,7 +608,7 @@ static Package *parse_manifest (Options *op)
             }
             free(tmpstr);
             
-            /* now, parse the flags */
+            /* every file has a type field */
 
             p->entries[n].flags = 0x0;
 
@@ -511,42 +623,81 @@ static Package *parse_manifest (Options *op)
                 p->entries[n].flags |= FILE_TYPE_OPENGL_HEADER;
             else if (strcmp(flag, "OPENGL_LIB") == 0)
                 p->entries[n].flags |= FILE_TYPE_OPENGL_LIB;
-            else if (strcmp(flag, "XFREE86_LIB") == 0)
-                p->entries[n].flags |= FILE_TYPE_XFREE86_LIB;
+            else if (strcmp(flag, "LIBGL_LA") == 0)
+                p->entries[n].flags |= FILE_TYPE_LIBGL_LA;
+            else if (strcmp(flag, "XLIB_STATIC_LIB") == 0)
+                p->entries[n].flags |= FILE_TYPE_XLIB_STATIC_LIB;
+            else if (strcmp(flag, "XLIB_SHARED_LIB") == 0)
+                p->entries[n].flags |= FILE_TYPE_XLIB_SHARED_LIB;
+            else if (strcmp(flag, "TLS_LIB") == 0)
+                p->entries[n].flags |= FILE_TYPE_TLS_LIB;
+            else if (strcmp(flag, "UTILITY_LIB") == 0)
+                p->entries[n].flags |= FILE_TYPE_UTILITY_LIB;
             else if (strcmp(flag, "DOCUMENTATION") == 0)
                 p->entries[n].flags |= FILE_TYPE_DOCUMENTATION;
+            else if (strcmp(flag, "MANPAGE") == 0)
+                p->entries[n].flags |= FILE_TYPE_MANPAGE;
             else if (strcmp(flag, "OPENGL_SYMLINK") == 0)
                 p->entries[n].flags |= FILE_TYPE_OPENGL_SYMLINK;
-            else if (strcmp(flag, "XFREE86_SYMLINK") == 0)
-                p->entries[n].flags |= FILE_TYPE_XFREE86_SYMLINK;
+            else if (strcmp(flag, "XLIB_SYMLINK") == 0)
+                p->entries[n].flags |= FILE_TYPE_XLIB_SYMLINK;
+            else if (strcmp(flag, "TLS_SYMLINK") == 0)
+                p->entries[n].flags |= FILE_TYPE_TLS_SYMLINK;
+            else if (strcmp(flag, "UTILITY_SYMLINK") == 0)
+                p->entries[n].flags |= FILE_TYPE_UTILITY_SYMLINK;
             else if (strcmp(flag, "INSTALLER_BINARY") == 0)
                 p->entries[n].flags |= FILE_TYPE_INSTALLER_BINARY;
-            else if (strcmp(flag, "NVIDIA_SETTINGS_BINARY") == 0)
-                p->entries[n].flags |= FILE_TYPE_NVIDIA_SETTINGS_BINARY;
-            else if (strcmp(flag, "OPENGL_LIB_CLASSIC_TLS") == 0)
-                p->entries[n].flags |= (FILE_TYPE_OPENGL_LIB |
-                                        FILE_CLASS_CLASSIC_TLS);
-            else if (strcmp(flag, "OPENGL_LIB_NEW_TLS") == 0)
-                p->entries[n].flags |= (FILE_TYPE_OPENGL_LIB |
-                                        FILE_CLASS_NEW_TLS);
-            else if (strcmp(flag, "XFREE86_LIB_CLASSIC_TLS") == 0)
-                p->entries[n].flags |= (FILE_TYPE_XFREE86_LIB |
-                                        FILE_CLASS_CLASSIC_TLS);
-            else if (strcmp(flag, "XFREE86_LIB_NEW_TLS") == 0)
-                p->entries[n].flags |= (FILE_TYPE_XFREE86_LIB |
-                                        FILE_CLASS_NEW_TLS);
-            else if (strcmp(flag, "OPENGL_SYMLINK_CLASSIC_TLS") == 0)
-                p->entries[n].flags |= (FILE_TYPE_OPENGL_SYMLINK |
-                                        FILE_CLASS_CLASSIC_TLS);
-            else if (strcmp(flag, "OPENGL_SYMLINK_NEW_TLS") == 0)
-                p->entries[n].flags |= (FILE_TYPE_OPENGL_SYMLINK |
-                                        FILE_CLASS_NEW_TLS);
-
-            else
+            else if (strcmp(flag, "UTILITY_BINARY") == 0)
+                p->entries[n].flags |= FILE_TYPE_UTILITY_BINARY;
+            else if (strcmp(flag, "DOT_DESKTOP") == 0)
+                p->entries[n].flags |= FILE_TYPE_DOT_DESKTOP;
+            else if (strcmp(flag, "XMODULE_SHARED_LIB") == 0)
+                p->entries[n].flags |= FILE_TYPE_XMODULE_SHARED_LIB;
+            else if (strcmp(flag, "XMODULE_SYMLINK") == 0)
+                p->entries[n].flags |= FILE_TYPE_XMODULE_SYMLINK;
+            else {
+                nvfree(flag);
                 goto invalid_manifest_file;
-            
-            free(flag);
-            
+            }
+
+            nvfree(flag);
+
+            /* some libs/symlinks have an arch field */
+
+            if (p->entries[n].flags & FILE_TYPE_HAVE_ARCH) {
+                flag = read_next_word(c, &c);
+                if (!flag) goto invalid_manifest_file;
+
+                if (strcmp(flag, "COMPAT32") == 0)
+                    p->entries[n].flags |= FILE_CLASS_COMPAT32;
+                else if (strcmp(flag, "NATIVE") == 0)
+                    p->entries[n].flags |= FILE_CLASS_NATIVE;
+                else {
+                    nvfree(flag);
+                    goto invalid_manifest_file;
+                }
+
+                nvfree(flag);
+            }
+
+            /* some libs/symlinks have a class field */
+
+            if (p->entries[n].flags & FILE_TYPE_HAVE_CLASS) {
+                flag = read_next_word(c, &c);
+                if (!flag) goto invalid_manifest_file;
+
+                if (strcmp(flag, "CLASSIC") == 0)
+                    p->entries[n].flags |= FILE_CLASS_CLASSIC_TLS;
+                else if (strcmp(flag, "NEW") == 0)
+                    p->entries[n].flags |= FILE_CLASS_NEW_TLS;
+                else {
+                    nvfree(flag);
+                    goto invalid_manifest_file;
+                }
+
+                nvfree(flag);
+            }
+
             /* libs and documentation have a path field */
 
             if (p->entries[n].flags & FILE_TYPE_HAVE_PATH) {
@@ -585,7 +736,7 @@ static Package *parse_manifest (Options *op)
 
     } while (!done);
     
-    if (manifest) munmap(manifest, len);
+    munmap(manifest, len);
     if (fd != -1) close(fd);
 
     return p;
@@ -603,7 +754,7 @@ static Package *parse_manifest (Options *op)
  fail:
     if (p && p->entries) free(p->entries);
     if (p) free(p);
-    if (manifest) munmap(manifest, len);
+    if (manifest != MAP_FAILED) munmap(manifest, len);
     if (fd != -1) close(fd);
     return NULL;
        
